<?php
/*
Plugin Name: Google Latitude History
Plugin URI: http://worldtravelblog.com/code/google-latitude-history-plugin/
Description: This plugin allows you to display your Google Latitude location history on Google Maps
Version: 0.9.8
Author: Peter Rosanelli
Author URI: http://www.worldtravelblog.com
*/

/**
* LICENSE
* This file is part of Google Latitude History Plugin.
*
* Google Latitude History Plugin is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
* @package    google-latitude-history
* @author     Peter Rosanelli <peter@worldtravelblog.com>
* @copyright  Copyright 2011 Peter Rosanelli
* @license    http://www.gnu.org/licenses/gpl.txt GPL 2.0
* @version    0.9.8
* @link       http://worldtravelblog.com/code/google-latitude-history-plugin/
*/


@session_start();
require_once 'google-api-php-client/src/apiClient.php';
require_once 'google-api-php-client/src/contrib/apiLatitudeService.php';

global $wpdb;

// settings page
require_once('settings.php');
// generates map from shortcode
require_once('google-latitude-history-map.php');

if(class_exists('GoogleLatitudeHistory')) {
	$plugin_glatitudehistory = new GoogleLatitudeHistory();
}

if(isset($plugin_glatitudehistory)) {
	/**
	 * Creates plugin option page
	 * This is outside of the class b/c syncLocations()
	 * which is called from options page will fail. it fails
	 * b/c $this is called outside of the object context. 
	 */
	function glatitudehistory_admin_menu() {
		global $plugin_glatitudehistory;
		add_options_page('Google Latitude History Settings', 'Google Latitude History', 'manage_options', 'google-latitude-history-menu', array(&$plugin_glatitudehistory, 'adminMenuOptions' ) );
	}
	
	register_activation_hook( __FILE__,  array(&$plugin_glatitudehistory, 'install' ));
	register_activation_hook(__FILE__, array(&$plugin_glatitudehistory, 'scheduleCron' ));
	register_deactivation_hook(__FILE__, array(&$plugin_glatitudehistory, 'unscheduleCron' ));
	register_deactivation_hook(__FILE__, array(&$plugin_glatitudehistory, 'dropDatebase' ));
	
	// shortcode scripts
	wp_register_script('google_maps', 'http://maps.googleapis.com/maps/api/js?sensor=false', false, '3', true);
	wp_register_script('styled-marker', plugins_url('js/StyledMarker.js', __FILE__), array('google_maps'), '0.5', true);
	wp_register_script('google-latitude-history', plugins_url('js/glatitudehistory.js', __FILE__), array('jquery', 'google_maps', 'styled-marker'), '0.9.8', true);
	// admin scripts
	wp_register_script('jquery-ui-datepicker', plugins_url('js/jquery.ui.datepicker.min.js', __FILE__), array('jquery-ui-core'), '1.8.16', true);
	wp_register_script('datetimepicker', plugins_url('js/datetimepicker.js', __FILE__), array('jquery-ui-datepicker'), '0.9.8', true);
	wp_register_script('jquery-ui-slider', plugins_url('js/jquery.ui.slider.min.js', __FILE__), array('jquery-ui-core', 'jquery-ui-mouse', 'jquery-ui-widget'), '1.8.16', true);
	wp_register_script('timepicker', plugins_url('js/timepicker.js', __FILE__), array('jquery-ui-datepicker', 'jquery-ui-slider'), '0.9.8', true);
	// admin styles
	wp_register_style('admin', plugins_url('css/admin.css', __FILE__));
	wp_register_style('jquery-ui-redmond', plugins_url('css/jquery-ui-1.8.16.redmond.css', __FILE__));	
	
	add_action('admin_menu', 'glatitudehistory_admin_menu');
	add_action('admin_print_scripts-settings_page_google-latitude-history-menu', array(&$plugin_glatitudehistory, 'adminPrintScripts'));
	add_action('admin_print_styles-settings_page_google-latitude-history-menu', array(&$plugin_glatitudehistory, 'adminPrintStyles'));
	add_action('hourly_latitude_sync', array(&$plugin_glatitudehistory, 'syncLocations'));
	add_action('admin_notices', array(&$plugin_glatitudehistory, 'oauthCheck'));	
	add_action('wp_footer', array(&$plugin_glatitudehistory, 'printScript'));
	add_shortcode('glatitude-map', array(&$plugin_glatitudehistory, 'shortcodeHandler'));
}

class GoogleLatitudeHistory {
	
	const DEFAULT_MAX_LOCATIONS = 20;
	const MAX_LOCATIONS = 100;
	const DEFAULT_CSS = 'margin:0 auto;font-size:12px;line-height:16px;';
	const LATITUDE_TABLE = 'glatitude_history'; 
	const CACHE_TABLE = 'api_cache';
	const RESPONSE_FIELDS = 'items(latitude,longitude,timestampMs,accuracy)'; // restrict response to relevant fields
	
	// determines if shortcode has been used on a page, if so then print the associated javascript
	static $add_script;
	
	// for counting the number of map instances on a single page and gives the data different ids
	static $instanceCount = 0;
	
	public $oauthClient;
	public $oauthService;
	public $redirectUrl;
	
	function GoogleLatitudeHistory() {
		
		$this->redirectUrl = get_admin_url() . 'options-general.php?page=google-latitude-history-menu';

		$client = new apiClient();
		$client->setClientId(get_option('glatitudehistory_client_id'));
		$client->setClientSecret(get_option('glatitudehistory_client_secret'));
		$client->setRedirectUri($this->redirectUrl);
		$client->setApplicationName(get_bloginfo('name'));
		if(get_option('glatitudehistory_access_token') != null) {
			$client->setAccessToken(get_option('glatitudehistory_access_token'));
		}
		$this->oauthClient = $client;
		$service = new apiLatitudeService($client);
		$this->oauthService = $service;
		
		if (isset($_GET['code'])) {
			try {
				$client->authenticate();
  				update_option('glatitudehistory_access_token', $client->getAccessToken());
  				header('Location: ' . $_SERVER['SCRIPT_NAME'] . '?page=google-latitude-history-menu');
			} catch (Exception $e) {
				//throw $e;
				
				//echo '<div class="error"><p>Your Client Secret is incorrect, please check the Google API Console. Reason: '.$e->getMessage().'</p></div>';
			}
		}
	}
	
	/**
	* Install Google Latitude History Plugin.
	* Register settings, set defaults, and create database table
	*/
	function install() {
		
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_google_api_key' );
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_google_api_ssl' );
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_use_google_api_key' );
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_client_id' );
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_client_secret' );
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_access_token' );
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_granularity' );
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_max_locations' );
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_location' );
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_custom_css' );
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_min_time' );
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_max_time' );
		register_setting( 'glatitudehistory-settings-group', 'glatitudehistory_latitude_api_key' );
		
		// default latitude query parameter for location history
		update_option('glatitudehistory_location', 'all'); 
		// default maximum number of locations
		update_option('glatitudehistory_max_locations', self::DEFAULT_MAX_LOCATIONS);
		// default css for map div
		update_option('glatitudehistory_custom_css', self::DEFAULT_CSS);
				
		$this->createTables();
	}
	
	/**
	* Creates the database tables used to store Google Latitude locations
	*/
	function createTables() {
		global $wpdb;
		
		// create table to store google latitude locations
		$sql = "DROP TABLE IF EXISTS ". $wpdb->prefix.self::LATITUDE_TABLE;
		$wpdb->query($sql);
		
   		if($wpdb->get_var("show tables like '".$wpdb->prefix.self::LATITUDE_TABLE."'") != $wpdb->prefix.self::LATITUDE_TABLE) {

   			$sql = "CREATE TABLE " . $wpdb->prefix.self::LATITUDE_TABLE . " (
   				id INT AUTO_INCREMENT PRIMARY KEY,
   				timestamp_ms BIGINT(20) NOT NULL,
   				date VARCHAR(200) NOT NULL,
   				latitude decimal(12,9) NOT NULL,
   				longitude decimal(12,9) NOT NULL,
   				accuracy INT(11) NOT NULL,
   				location_name VARCHAR(512) default NULL,
   				INDEX idx_timestamp_ms (timestamp_ms));";
   			
   			require_once(ABSPATH . 'wp-admin/includes/upgrade.php');
			dbDelta($sql);
   		}
   		
   		// create table for google api php client to store cache
		$sql = "DROP TABLE IF EXISTS ". $wpdb->prefix.self::CACHE_TABLE;
		$wpdb->query($sql);
		
   		if($wpdb->get_var("show tables like '".$wpdb->prefix.self::CACHE_TABLE."'") != $wpdb->prefix.self::CACHE_TABLE) {

   			$sql = "CREATE TABLE " . $wpdb->prefix.self::CACHE_TABLE . " (
 				id INT NOT NULL AUTO_INCREMENT PRIMARY KEY, 
 				k VARCHAR(32),  
 				data TEXT, 
 				last_updated TIMESTAMP DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP,
   				INDEX idx_key (k));";
   			
   			require_once(ABSPATH . 'wp-admin/includes/upgrade.php');
			dbDelta($sql);
   		}
	}
	
	/**
	* Drops the table used to store Google Latitude locations
	*/
	function dropTable() {
		global $wpdb;
		
		$sql = "DROP TABLE IF EXISTS ". $wpdb->prefix.self::LATITUDE_TABLE;
		
		$wpdb->query($sql);
	}
	
	/**
	* Creates sync location cron job
	*/
	function scheduleCron() {
		wp_schedule_event(time(), 'hourly', 'hourly_latitude_sync' );
	}
	
	/**
	* Removes sync location cron job
	*/
	function unscheduleCron() {
		wp_clear_scheduled_hook( 'hourly_latitude_sync' );
	}
		
	/**
	* Checks if the OAuth token is stale by using it to call
	* the Google Latitude rest service for the current location
	*
	* @return string error message for stale token
	*/
	function oauthCheck() {
		
		// get current location
		try {
			$response = $this->oauthService->currentLocation->get();
		} catch (Exception $e) {
			echo '<div class="error"><p><span style="font-weight:bold">Google Latitude History: </span>Your OAuth token is stale, please refresh the token on the <a href="options-general.php?page=google-latitude-history-menu">settings page</a>. Reason: '.$e->getMessage().'</p></div>';
		}
	}
	
	/**
	 * Scripts printed on the Google Latitude History admin page
	 */
	function adminPrintScripts() {
		wp_enqueue_script('jquery-ui-datepicker');
		wp_enqueue_script('datetimepicker');
		wp_enqueue_script('jquery-ui-slider');
		wp_enqueue_script('timepicker');
	}
	
	/**
	 * Styles printed on the Google Latitude History admin page
	 */
	function adminPrintStyles() {
		wp_enqueue_style('admin');
		wp_enqueue_style('jquery-ui-redmond');
	}
	
	/**
	*  add google latitude history link to the admin menu
	*/
	function adminMenu() {
		add_options_page('Google Latitude History Settings', 'Google Latitude History', 'manage_options', 'google-latitude-history-menu', array('GoogleLatitudeHistory', 'adminMenuOptions' ) );
	}
	
	/**
	* This function handles the plugin admin settings page which includes
	* updating the settings, syncing the database with Google Latitude, and
	* displaying the settings page 
	*
	* @return string the settings page as html
	*/
	function adminMenuOptions() {
		global $wpdb;
		
		echo( '<h2>Google Latitude History Settings</h2>' );
		
			wp_nonce_field( 'update-options' );

			if(isset($_POST['submit']) && ($_POST['submit'] == 'Update Options' || $_POST['submit'] == 'Refresh OAuth Token')) {

				// Update Google Latitude Settings
				update_option('glatitudehistory_google_api_key', $_POST['google_api_key']);				
				
				if(isset($_POST['google_api_ssl']) && $_POST['google_api_ssl'] == 'true') {
					update_option('glatitudehistory_google_api_ssl', $_POST['google_api_ssl']);	
				} else {
					update_option('glatitudehistory_google_api_ssl', '');
				}
				
				if(isset($_POST['use_google_api_key']) && $_POST['use_google_api_key'] == 'true') {
					update_option('glatitudehistory_use_google_api_key', $_POST['use_google_api_key']);	
				} else {
					update_option('glatitudehistory_use_google_api_key', '');
				}
				
				update_option('glatitudehistory_client_id', $_POST['client_id']);
				update_option('glatitudehistory_client_secret', $_POST['client_secret']);
				update_option('glatitudehistory_granularity', $_POST['granularity']);
				update_option('glatitudehistory_custom_css', $_POST['custom_css']);
				
				// check if max locations parameter is a number and gt 0 and lt MAX_LOCATIONS
				// if not then set it to the default number of locations
				update_option('glatitudehistory_max_locations', self::DEFAULT_MAX_LOCATIONS);
				if( is_numeric($_POST['max_locations']) ) {
					$max_locations = (int) $_POST['max_locations'];
					if($max_locations > 0 && $max_locations < self::MAX_LOCATIONS) {
						update_option( 'glatitudehistory_max_locations', $max_locations);
					}
				}
				
				update_option('glatitudehistory_min_time', $_POST['min_time']);
				update_option('glatitudehistory_max_time', $_POST['max_time']);
					
				echo '<div class="updated"><p>Settings Updated</p></div>';
				
//				if($_POST['submit'] == 'Refresh OAuth Token') {
//					header('Location: ' . $this->oauthClient->createAuthUrl());
//				}
			
			} elseif(isset($_POST['submit']) && $_POST['submit'] == 'Sync Location History' ) {
				
				$statusMessage = '';
				
				try { 
					$latitudeResponse = $this->syncLocations();
					
					$statusMessage = '<div class="updated"><p>Google Latitude History was synced <br/>';
					$statusMessage .= '<a href="#" id="toggle_response">see latitude response</a>';
					$statusMessage .= '<div id="json_response" style="display:none">'.json_encode($latitudeResponse).'</div>';
					$statusMessage .= '</p></div>';					
					
				} catch (Exception $e) {
					
					$statusMessage .= '<div class="error"><p>Error syncing Latitude location history: '.$e->getMessage();
					$statusMessage .= '</p></div>';
				}
				
				// sync status message
				echo $statusMessage;
			}
			
			$this->oauthClient->setClientId(get_option('glatitudehistory_client_id'));
			$this->oauthClient->setClientSecret(get_option('glatitudehistory_client_secret'));
					
			// display the settings page
			glatitudehistory_settings($this->oauthClient->createAuthUrl(), $this->redirectUrl);
	}
	
	/**
	* Calls the Google Latitude rest service to get the location history and 
	* updates the database.
	* 
	* @return object latitude response
	*/
	function syncLocations() {
		
		$options = array();
		
		$options['max-results'] = get_option('glatitudehistory_max_locations');
		
		// min/max-time is milliseconds since the epoch hence multipling by 1000
		if(get_option('glatitudehistory_min_time')) {
			$options['min-time'] = strtotime(get_option('glatitudehistory_min_time'))*1000;
		}
		if(get_option('glatitudehistory_max_time')) {
			$options['max-time'] = strtotime(get_option('glatitudehistory_max_time'))*1000;
		}
		
		$options['granularity'] = get_option('glatitudehistory_granularity');
		
		// restrict response to relevant fields
		$options['fields'] = self::RESPONSE_FIELDS;

		// get locations from glatitude
		$latitudeResponse = $this->oauthService->location->listLocation($options);
		// update database
		$this->updateDatabase($latitudeResponse);

		return $latitudeResponse;
	}
	
	/**
	* Takes the location history and compares it to whats in the 
	* the database then adds new locations and removes old ones
	* from the database
	*
	* @param object location history
	*/
	function updateDatabase($latitudeHistory) {
		global $wpdb;
	
		// create hash of location info by timestamp from glatitude
		// and put the glatitude locations (new locations) in an array for comparison to the database locaations
		$googleTimestamps = array();
		$googleLocationsByTimestamp = array();
		if($latitudeHistory->getItems()) {
			foreach($latitudeHistory->getItems() as $location) {
				if($location->getTimestampMs()) {
					$googleTimestamps[] = $location->getTimestampMs();
					$googleLocationsByTimestamp[$location->getTimestampMs()] = $location; 
				} else {
					throw new Exception('Empty Google Latitude location data. <a href="http://code.google.com/p/latitude-api/issues/detail?id=30" target="_blank">bug report</a>');
				}
			}
		} else {
			throw new Exception('No Google Latitude locations found.');
		}
			
		// put the database locataions ("old locations") into an array for comparison to the glatitude locations
		$wpHistory = $wpdb->get_results('SELECT * FROM ' . $wpdb->prefix.self::LATITUDE_TABLE);
		$wpTimestamps = array();
		foreach($wpHistory as $location) {
			$wpTimestamps[] = $location->timestamp_ms;
		}
		
		// find new locations from glatitude
		$newTimestamps = array_diff($googleTimestamps, $wpTimestamps);
	
		// find old locations in the database
		$oldTimestamps = array_diff($wpTimestamps, $googleTimestamps);
			
		// insert new locations into database
		foreach($newTimestamps as $timestamp) {
			$location = $googleLocationsByTimestamp[$timestamp];
			
			// get location name by reverse code
			$locationName = $this->getLocationName($location->getLatitude(), $location->getLongitude());
			
			$wpdb->insert( $wpdb->prefix.self::LATITUDE_TABLE, array(
				'timestamp_ms' => $location->getTimestampMs(), 
				'date' => date("F j, Y, g:i a e", $location->getTimestampMs()/1000), // format date to "July 7, 2011, 12:58 pm UTC"
				'latitude' => $location->getLatitude(),
				'longitude' => $location->getLongitude(),
				'accuracy' => $location->getAccuracy(),
				'location_name' => $locationName
			), array( '%s', '%s', '%s', '%s', '%d', '%s' ) );
		}
			
		// remove old locations from database
		if(count($oldTimestamps) > 0) {
			$wpdb->query('DELETE FROM '.$wpdb->prefix.self::LATITUDE_TABLE.' WHERE timestamp_ms in ('.join(',', $oldTimestamps). ');');
		}
	}
	
	/**
	* Takes location coordinates and uses Google reverse geocoding service
	* to get the location name
	*
	* @param string latitude
	* @param string longitude
	* @return string location name
	*/
	function getLocationName($lat, $lng) {
		
		// reverse geocode the coordinates to get the location name
		$response = json_decode(file_get_contents('http://maps.googleapis.com/maps/api/geocode/json?sensor=false&latlng='.$lat.','.$lng));
		
		// go through the results to find the formatted address (there are several options)
		$address = array();
		if($response->status == 'OK') {
			foreach($response->results as $result) {
				foreach($result->types as $type) {
					if($type == 'locality') {
						$address['locality'] = $result->formatted_address;
						break;
					} elseif($type == 'administrative_area_level_2') {
						$address['administrative_area_level_2'] = $result->formatted_address;
						break;
					} elseif($type == 'administrative_area_level_1') {
						$address['administrative_area_level_1'] = $result->formatted_address;
						break;
					} elseif($type == 'postal_code') {
						$address['postal_code'] = $result->formatted_address;
						break;
					}
				}
				// most preferred formatted address so let's stop
				if(isset($address['locality'])) {
					break;
				}
			}
		}
		
		// Prefer for formatted address = locality > administrative_area_level_2 > administrative_area_level_1 > postal_code
		$locationName = "";
		if(isset($address['locality'])) {
			$locationName = $address['locality'];
		} elseif(isset($address['administrative_area_level_2'])) {
			$locationName = $address['administrative_area_level_2'];
		} elseif(isset($address['administrative_area_level_1'])) {
			$locationName = $address['administrative_area_level_1'];
		} elseif(isset($address['postal_code'])) {
			$locationName = $address['postal_code'];
		}
	
		return $locationName;
	}
	
	/**
	* The function turns the shortcode into html and javascript for displaying the
	* Google Latitude location history on a Google Map
	*
	* @param array shortcode attributes
	* @return string html and javascript for map
	*/
	function shortcodeHandler($attrs) {
		global $wpdb;
		
		self::$add_script = true;
		
		self::$instanceCount++;
		
		// add min/max time condition
		$condition = ' WHERE ';
		$timeConditionSet = false;
		if(isset($attrs['mintime']) && $attrs['mintime'] != '') {
			$minTime = strtotime($attrs['mintime'])*1000;
			if($minTime) {
				$condition .= ' timestamp_ms >= '.$minTime;
				$timeConditionSet = true;
			}
		}
		if(isset($attrs['maxtime']) && $attrs['maxtime'] != '') {
			$maxTime = strtotime($attrs['maxtime'])*1000;
			if($maxTime) {
				if($timeConditionSet) {
					$condition .= ' and ';
				}
				$condition .= ' timestamp_ms <= '.$maxTime;
				$timeConditionSet = true;	
			}
		}
		
		// last location is the first result
		$query = 'SELECT * FROM '.$wpdb->prefix.self::LATITUDE_TABLE.' ';
		if($timeConditionSet) {
			$query .= $condition;
		}
		$query .= ' ORDER BY timestamp_ms DESC';
		
		$latitudeHistory = $wpdb->get_results($query);
	
		return GoogleLatitudeHistoryMap::generateMap($attrs, $latitudeHistory, get_option('glatitudehistory_custom_css'), self::$instanceCount);
	}
	
	/**
	* Decides whether the javascript files with be included in a page
	*/
	function printScript() {
		if ( ! self::$add_script )
		return;
		
		// 
		if(get_option('glatitudehistory_google_api_ssl') == 'true' || get_option('glatitudehistory_use_google_api_key') == 'true') {

			wp_deregister_script('google_maps');
			
			$protocol = 'http';
			if(get_option('glatitudehistory_google_api_ssl') == 'true') {
				$protocol = 'https';								
			}
			
			$urlKey = '';
			if(get_option('glatitudehistory_use_google_api_key') == 'true') {
				$urlKey = '&key='.get_option('glatitudehistory_google_api_key');
			}
			
			$googleMapApiUrl = $protocol.'://maps.googleapis.com/maps/api/js?sensor=false'.$urlKey;
			
			wp_register_script('google_maps', $googleMapApiUrl, false, '3', true);
		}
		
		// used print instead of enqueue b/c it doesnt work when printing the js on the bases of a variable 
		wp_print_scripts('google_maps');
		wp_print_scripts('styled-marker');
		wp_print_scripts('google-latitude-history');
	}
	
	
}	

?>